import datetime
import os

from horno.datos.Encoding import Encoding
from horno.datos.TiposDeDatos import DatoStr
from horno.datos.bd import dbfonpy
from horno.utiles.CSV import CSVEscritor
from horno.utiles.IO import IOSistema
from horno.utiles.Metricas import Progreso
import xlrd

import pyodbc


#=============================================================================================
class CxBase (object):

    #------------------------------------------------------------------------------------------
    def __init__(self):
        
        ''

    #------------------------------------------------------------------------------------------
    def Conexion(self):

        return None
    
    #------------------------------------------------------------------------------------------
    def Cursor(self, conexion):

        return None

    #------------------------------------------------------------------------------------------
    def FuentesODBC(self, nombre):

        return pyodbc.dataSources().get(nombre, '')

    #------------------------------------------------------------------------------------------
    def RealizarQueries(self, queries):
        
        try:
            conn = self.Conexion()
            curs = self.Cursor(conn)

            progreso = Progreso(1, len(queries), 'Ejecutando Querys')
            
            for query in queries:
                
                curs.execute(query)
                
                if query.strip().upper().startswith('SELECT'): 
                    datos = curs.fetchall()
                    columnas = Encoding().NormalizarLista([ d[0] for d in curs.description ])
                    conn.commit()
                else:
                    datos = [];
                    columnas = [];
                    conn.commit()
                    
                progreso.Incrementar()
                
            curs.close()
            conn.close()
            
            return [columnas, datos, None]
            
        except Exception as e:
            return [None, None, '[EXCEPTION] ' + str(e)]

    #------------------------------------------------------------------------------------------
    def RealizarQuery(self, query):
        
        return self.RealizarQueries([query])

    #------------------------------------------------------------------------------------------
    def ArmarSelect(self, tabla):
        
        return 'SELECT * FROM [%s]' % (tabla)

    #------------------------------------------------------------------------------------------
    def ExportarCSV(self, archivo_csv_output, tabla='', gen_script=False, archivo_sql_output='', append=False):
        
        query_select = self.ArmarSelect(tabla) 
        [columnas, datos, result] = self.RealizarQuery(query_select)

        if columnas is None or datos is None:
            IOSistema().PrintLine('[ConexionBases] Warning: (%s) El query no arrojo ningun resultado' % (archivo_csv_output))
            IOSistema().PrintLine('[Result]: %s' % (result))
            columnas = []
            datos = []

        headerstruc = [DatoStr(c,True) for c in columnas]
        csvw = CSVEscritor(archivo_csv_output, headerstruc)
        csvw.EscribirDatos(datos)
        
        def format_comillas(nombre):
            if ' ' in nombre or '/' in nombre or nombre[0] in '1234567890-' or nombre.lower() in ['any']:
                return '"%s"' % (nombre)
            else:
                return nombre
        
        if gen_script:
            com_tabla = format_comillas(tabla) 
            if archivo_sql_output=='':
                archivo_sql_output = '%s.sql' % (archivo_csv_output)
            f = open(archivo_sql_output, 'a+' if append else 'w+')
            f.write('\n---------- %s ----------";' % (com_tabla))
            f.write('\nDROP TABLE IF EXISTS %s;' % (com_tabla))
            f.write('\nCREATE TABLE %s ();' % (com_tabla))
            for columna in columnas:
                com_columna = format_comillas(columna)
                f.write('\nALTER TABLE %s ADD COLUMN %s VARCHAR;' % (com_tabla, com_columna))
            f.write('\nSET CLIENT_ENCODING=\'LATIN1\';')
            f.write('\nCOPY %s FROM \'%s\' WITH DELIMITER \',\' CSV HEADER;' % (com_tabla, os.path.abspath(archivo_csv_output).replace('\\','/')))
            f.close()
        

#=============================================================================================
class CxBaseAccess (CxBase):

    #------------------------------------------------------------------------------------------
    def __init__(self, archivo_mdb_input):
        
        self._archivo = archivo_mdb_input

    #------------------------------------------------------------------------------------------
    def Conexion(self):

        cs = 'DRIVER=%s;DBQ=%s;PWD=%s' % (self.FuentesODBC('MS Access Database'), self._archivo, '')
        return pyodbc.connect(cs)
    
    #------------------------------------------------------------------------------------------
    def Cursor(self, conexion):

        return conexion.cursor()

    #------------------------------------------------------------------------------------------
    def RealizarQueries(self, queries):
    
        res = super(CxBaseAccess, self).RealizarQueries(queries)
        
        return res
    

#=============================================================================================
class CxBaseDBF (CxBase):

    #------------------------------------------------------------------------------------------
    def __init__(self, archivo_dbf_input):
        
        self._archivo = archivo_dbf_input

    #------------------------------------------------------------------------------------------
    def Conexion(self):

        return dbfonpy.connect(self._archivo)
    
    #------------------------------------------------------------------------------------------
    def Cursor(self, conexion):

        return conexion.cursor()
    
    #------------------------------------------------------------------------------------------
    def ExportarCSV(self, archivo_csv_output, tabla=''):
        
        CxBase.ExportarCSV(self, archivo_csv_output, 'dbf')
    
#=============================================================================================
class CxBaseExcel (CxBase):

    #------------------------------------------------------------------------------------------
    def __init__(self, archivo_xls_input, con_header=True):
        
        self._archivo = archivo_xls_input
        self._con_header = con_header

    #------------------------------------------------------------------------------------------
    def Conexion(self):

        cs = 'DRIVER=%s;DBQ=%s;PWD=%s' % (self.FuentesODBC('Excel Files'), self._archivo, '')
        return pyodbc.connect(cs, autocommit=True)
    
    #------------------------------------------------------------------------------------------
    def Cursor(self, conexion):

        return conexion.cursor()
    
    #------------------------------------------------------------------------------------------
    def ExportarCSV(self, archivo_csv_output, tabla=''):
        
        self._hoja_actual = tabla

        CxBase.ExportarCSV(self, archivo_csv_output, '"%s$"' % (tabla))
    
    #------------------------------------------------------------------------------------------
    def RealizarQuery(self, query, tabla=''):
    
        try:
            
            if tabla: self._hoja_actual = tabla
            
            # self.FuentesODBC('Excel Files')
            
            def tran(wb, sh, rownum):
                res = []
                for cell in sh.row(rownum):
                    try:
                        if cell.ctype == xlrd.XL_CELL_DATE:
                            val = datetime.datetime( *xlrd.xldate_as_tuple(cell.value, wb.datemode) )
                        elif cell.ctype == xlrd.XL_CELL_NUMBER and cell.value.is_integer():
                            val = int(cell.value)
                        elif cell.ctype == xlrd.XL_CELL_TEXT:
                            val = Encoding().NormalizarTexto(cell.value)
                        else:
                            val = cell.value
                    except Exception as e2:
                        val = '<ERROR_XLS>'
                    finally:
                        res.append(val)
                return res 
                
            wb = xlrd.open_workbook(self._archivo.Ruta())
            sh = wb.sheet_by_name(self._hoja_actual)

            columnas = []
            datos = []
            for rownum in range(sh.nrows):
                if rownum == 0 and self._con_header:
                    columnas = tran(wb, sh, rownum)
                else:
                    datos.append(tran(wb, sh, rownum))
        
            return [columnas, datos, '']
            
        except Exception as e:
            print 
            return [None, None, '[EXCEPTION] ' + str(e)]



import re

from PyQt4 import QtGui, QtCore
from horno.datos.Encoding import Encoding
from horno.utiles.IO import IOSistema


#==============================================================================================================================
class VistaGraficaBase (QtGui.QWidget):
    
    #------------------------------------------------------------------------------------------
    def __init__(self, controlador, args=[]):

        self._controlador = controlador
        self._args = args
        self._app = QtGui.QApplication(self._args)
        self._helper = VistaGraficaHelper(self)
        QtGui.QWidget.__init__(self)

    #------------------------------------------------------------------------------------------
    def Mostrar(self):
        
        self._app.setActiveWindow(self)
        self.show()
        self._app.exec_()

    #---------------------------------------------------------------------------------------------------------
    def Notificar(self, signal, pars=None):

        self.emit(QtCore.SIGNAL(signal), *pars)

    #---------------------------------------------------------------------------------------------------------
    def ConectarEvento(self, signal, callback):

        self.connect(self, QtCore.SIGNAL(signal), callback)
            

#==============================================================================================================================
class VistaGraficaHelper:
    
    #------------------------------------------------------------------------------------------
    def __init__(self, padre):
        
        self._padre = padre

    #------------------------------------------------------------------------------------------
    def AgregarLayout(self, padre, horz):
        
        layout = QtGui.QHBoxLayout() if horz else QtGui.QVBoxLayout()
        try: 
            padre.addLayout(layout)
        except:
            padre.setLayout(layout)
            
        return layout

    #------------------------------------------------------------------------------------------
    def AgregarWidget(self, padre, widget, horz, texto):
        
        layout = self.AgregarLayout(padre, horz)

        if texto:
            lbl = QtGui.QLabel(texto)
            layout.addWidget(lbl)

        wgt = widget
        layout.addWidget(wgt)
        return [lbl, wgt]

    #------------------------------------------------------------------------------------------
    def AgregarTab(self, padre, texto, w=None, h=None):
        
        tab = QtGui.QWidget()
        if w: tab.setMaximumWidth(w)
        if h: tab.setMaximumHeight(h)
        padre.addTab(tab, texto)
        return tab

    #------------------------------------------------------------------------------------------
    def AgregarTexto(self, padre, horz, texto, readonly=False, w=None, h=None):

        layout = self.AgregarLayout(padre, horz)
        lbl = QtGui.QLabel(texto)
        layout.addWidget(lbl)

        txt = QtGui.QTextEdit()
        if w: txt.setFixedWidth(w)
        if h: txt.setFixedHeight(h)
        txt.setReadOnly(readonly)
        txt.setAlignment(QtCore.Qt.AlignTop)
        layout.addWidget(txt)
        return [lbl, txt]

    #------------------------------------------------------------------------------------------
    def AgregarWeb(self, padre, horz, texto, w=None, h=None):

        layout = self.AgregarLayout(padre, horz)
        lbl = QtGui.QLabel(texto)
        layout.addWidget(lbl)

        from PyQt4.QtWebKit import QWebView
        
        web = QWebView()
        if w: web.setFixedWidth(w)
        if h: web.setFixedHeight(h)
        layout.addWidget(web)
        return [lbl, web]

    #------------------------------------------------------------------------------------------
    def AgregarBoton(self, padre, texto, callback):

        btn = QtGui.QPushButton(texto)
        btn.clicked.connect(callback)
        padre.addWidget(btn)
        return btn

    #------------------------------------------------------------------------------------------
    def AgregarLista(self, padre, horz, texto, callback):
        
        layout = self.AgregarLayout(padre, horz)
        lbl = QtGui.QLabel(texto)
        layout.addWidget(lbl)
        
        lst = QtGui.QListWidget()
        lst.currentItemChanged.connect(callback)
        layout.addWidget(lst)
        return [lbl, lst]

    #------------------------------------------------------------------------------------------
    def AgregarTabla(self, padre, horz, texto):

        layout = self.AgregarLayout(padre, horz)
        lbl = QtGui.QLabel(texto)
        layout.addWidget(lbl)

        tbl = QtGui.QTableWidget()
        layout.addWidget(tbl)
        return [lbl, tbl]

    #------------------------------------------------------------------------------------------
    def CargarLista(self, lst, items):

        lst.clear()
        for item in items:
            lst.addItem(item)

    #------------------------------------------------------------------------------------------
    def CargarTabla(self, tbl, columnas, filas):

        if not columnas:
            columnas = ['(vacio)']

        tbl.setRowCount(len(filas))
        tbl.setColumnCount(len(columnas))
        tbl.setHorizontalHeaderLabels(columnas)
        for n, fila in enumerate(filas):
            for i, d in enumerate(fila):
                d_str = Encoding().ToUnicode(d, 'utf-8') if d else ''
                item = QtGui.QTableWidgetItem(d_str)
                item.setToolTip(d_str)
                tbl.setItem(n, i, item)
            
    #------------------------------------------------------------------------------------------
    def GuardarTabla(self, tbl, ruta_csv):
        
        sep = ','
        quo = '"'
        
        with open(ruta_csv, 'wb') as io:
            
            columnas = [self.AUnicode(tbl.horizontalHeaderItem(i).text()) for i in range(tbl.columnCount())]
            
            header = sep.join([quo + c.replace(quo, quo * 2) + quo for c in columnas])
            io.write(Encoding().ToString(header + '\n'))
            
            for f in range(tbl.rowCount()):
                fila = []
                for c in range(tbl.columnCount()):
                    item = tbl.item(f, c)
                    fila.append(self.AUnicode(item.text()) if item is not None else '')
                
                datos = sep.join([quo + d.replace(quo, quo * 2) + quo for d in fila])
                io.write(Encoding().ToString(datos + '\n'))
                                
    #------------------------------------------------------------------------------------------
    def ObtenerTabla(self, tbl):
        
        res = []
        
        for f in range(tbl.rowCount()):
            fila = []
            for c in range(tbl.columnCount()):
                item = tbl.item(f, c)
                fila.append(self.AUnicode(item.text()).strip() if item is not None else '')
            res.append(fila)
            
        return res
                                
    #------------------------------------------------------------------------------------------
    def ObtenerTexto(self, txt, selected=False, solo_ascii=False):

        valor = ''
        if selected: valor = self.AUnicode(txt.textCursor().selectedText())
        if not selected or not valor: valor = self.AUnicode(txt.toPlainText())
        if solo_ascii: valor = re.sub(r'[^\x00-\x7F]+',' ', valor)
        return valor

    #------------------------------------------------------------------------------------------
    def SetearTexto(self, ctl, valor):

        ctl.setText(valor)

    #------------------------------------------------------------------------------------------
    def SetearEstilo(self, ctl, estilo):

        ctl.setStyleSheet("QWidget {color:%s}" % (estilo))

    #------------------------------------------------------------------------------------------
    def SetearHabilitado(self, ctl, valor):

        ctl.setEnabled(valor)

    #------------------------------------------------------------------------------------------
    def ObtenerItem(self, lst):

        item = lst.currentItem()
        valor = str(item.text()) if item else ''
        return valor

    #------------------------------------------------------------------------------------------
    def AUnicode(self, txt):
        
        return txt # unicode(txt)
        

#==============================================================================================================================
class VistaConsolaBase (object):
    
    #------------------------------------------------------------------------------------------
    def __init__(self, controlador, args=[]):

        self._controlador = controlador
        self._args = args
        
    #------------------------------------------------------------------------------------------
    def Mostrar(self):
        
        IOSistema().PrintLine('Hola, soy una VistaConsolaBase y me estoy mostrando')




import json
from urllib.parse import parse_qs

from horno.utiles.IO import IOSistema
from horno.utiles.Singleton import Singleton
import unidecode


#=============================================================================================
class JsonHelper (metaclass=Singleton):

    #------------------------------------------------------------------------------------------
    def __init__(self):
        
        pass

    #------------------------------------------------------------------------------------------
    def find(self, d, path, default=''):
        keys = path.split('.')
        rv = d
        for key in keys:
            if not rv or key not in rv:
                return default
            rv = rv[key]
        return rv

    #------------------------------------------------------------------------------------------
    def set(self, d, path, valor):
        keys = path.split('.')
        rv = d
        for key in keys:
            if not rv or key not in rv:
                rv[key] = {}
            rv = rv[key]
        rv = valor
    
    #------------------------------------------------------------------------------------------
    def to_json(self, d, attr=None):

        try:

            if attr and hasattr(d, attr):
                d = getattr(d, attr)
            if not d:
                return {} 
            if isinstance(d, dict):
                return d
            if isinstance(d, str):
                return json.loads(d)
            if isinstance(d, bytes):
                return json.loads(d.decode('utf-8'))
            
            raise Exception('input invalido: %s' % (d))

        except Exception as exc:
            
            err = str(exc)
            IOSistema().PrintLine('(X) [to_json] ERROR: %s' % (err))
            return {}

    #------------------------------------------------------------------------------------------
    def to_json_from_qs(self, e):
        
        out_json = parse_qs(e)
        for (k, v) in out_json.items():
            out_json[k] = unidecode.unidecode(v[0] if v else '')
            
        return out_json

    #------------------------------------------------------------------------------------------
    def to_string(self, d, attr=None, ret=False):

        try:
            out_json = JsonHelper().to_json(d, attr)
            texto = json.dumps(out_json, indent=4, sort_keys=True)
        except Exception as exc:
            IOSistema().PrintLine('(X) [to_string] ERROR: %s' % (str(exc)))
            texto = str(d)
        finally:
            if ret:
                return texto
            else:
                IOSistema().PrintLine(texto)

    #------------------------------------------------------------------------------------------
    def to_markdown(self, d, ret=False, nivel=0, lista=False):

        try:
            texto = ''
            espacios_gap = 2
            if isinstance(d, str):
                texto += '%s%s\n' % (' ' * (nivel), d)
            elif isinstance(d, list):
                for e in d:
                    texto += JsonHelper().to_markdown(e, ret, nivel+espacios_gap, lista=True)
            elif isinstance(d, dict):
                for (dic_k, dic_v) in d.items():
                    espacios = len(dic_k) - len(dic_k.lstrip())
                    texto += '%s%s: %s' % (' ' * (nivel), dic_k, '' if isinstance(dic_v, str) else '\n')
                    texto += JsonHelper().to_markdown(dic_v, ret, (nivel+espacios)+espacios_gap)
        except Exception as exc:
            IOSistema().PrintLine('(X) [to_markdown] ERROR: %s' % (str(exc)))
        finally:
            if ret:
                return texto
            else:
                IOSistema().PrintLine(texto)



import logging

from horno.datos.Fechas import Fecha
from horno.utiles.IO import IOEscritor, IOArchivo, IOSistema, IOCarpeta


#================================================================================================
class Logger ():
    
    # ------------------------------------------------------------------------------------------
    def __init__(self, log_folder=None, data={}):

        self.sep_len = 60
        self.log_folder = log_folder
        self.printear = False
        self.config(data)

    # ------------------------------------------------------------------------------------------
    def config(self, data):

        self.printear = data.get('printear', False)
        self.timestamp = data.get('timestamp', False)

    # ------------------------------------------------------------------------------------------
    def get_log_file(self, log_file_rel):

        if self.log_folder:
            IOCarpeta(self.log_folder).Crear()
            log_file = '%s/%s' % (self.log_folder, log_file_rel)
        else:
            log_file = 'nul'

        return log_file

    # ------------------------------------------------------------------------------------------
    def reset(self, log_file_rel):

        log_file = self.get_log_file(log_file_rel)
        with IOEscritor.DeRuta(log_file).Abrir(False) as iow:
            iow.EscribirLinea('LOG FILE RESETEADO!')

    # ------------------------------------------------------------------------------------------
    def log(self, log_file_rel, texto):

        log_file = self.get_log_file(log_file_rel)
        with IOEscritor.DeRuta(log_file).Abrir(append=True, binario=False) as iow:
            if self.timestamp:
                fecha_ahora = Fecha.Ahora()
                sep = '-' * self.sep_len + '\n'
                iow.Escribir(sep + ('%s' % (fecha_ahora.to_human("%Y-%m-%d %H:%M:%S"))) + '\n')
            iow.Escribir(str(texto) + '\n')

        if self.printear:
            IOSistema().PrintLine(str(texto))

    # ------------------------------------------------------------------------------------------
    def attach(self, log_file_rel, clases):

        log_file = self.get_log_file(log_file_rel)
        for clase in clases:
            logger = logging.getLogger(clase)
            logger.handlers = [h for h in logger.handlers if not IOArchivo.PathFix(log_file) in IOArchivo.PathFix(h.baseFilename)]
            logger.setLevel(logging.DEBUG)
            handler = logging.FileHandler(log_file)
            handler.setLevel(logging.DEBUG)
            sep = '-' * self.sep_len + '\n'
            formatter = logging.Formatter(sep + '%(asctime)s, %(name)s [ %(levelname)s ]\n%(message)s' + '\n')
            handler.setFormatter(formatter)
            logger.addHandler(handler)



from horno.datos.Encoding import Encoding
from horno.datos.Registro import Registro
from horno.utiles.IO import IOLector, IOSistema
from horno.utiles.Metricas import Progreso
import csv

#=============================================================================================
class CSVLector:
    
    #------------------------------------------------------------------------------------------
    def __init__(self, nombre, archivo_csv_input, func_reg_id, delim=',', quote='"'):
        
        self._nombre = nombre
        self._archivo = archivo_csv_input
        self._delim = delim
        self._quote = quote
        self._func_reg_id = func_reg_id
        self._header = []
        self._datos = dict()
        self._inicializado = False

    #------------------------------------------------------------------------------------------
    def Nombre(self):
        return self._nombre

    #------------------------------------------------------------------------------------------
    def Archivo(self):
        return self._archivo

    #------------------------------------------------------------------------------------------
    def Header(self):
        return self._header

    #------------------------------------------------------------------------------------------
    def Inicializado(self):
        return self._inicializado

    #------------------------------------------------------------------------------------------
    def Claves(self):
        return self._datos.keys()

    #------------------------------------------------------------------------------------------
    def RegistroNuevo(self):
        reg = Registro()
        for h in self._header:
            reg.setv(h, '')
        return reg

    #------------------------------------------------------------------------------------------
    def ClaveExiste(self, key):
        return key in self._datos
            
    #------------------------------------------------------------------------------------------
    def RegistrosExisten(self, key):
        return self.ClaveExiste(key) and len(self._datos[key]) > 0

    #------------------------------------------------------------------------------------------
    def RegistrosPorId(self, key):
        if not self.ClaveExiste(key):
            return None

        regs = []
        for dato in self._datos[key]:
            reg = self.RegistroUsandoDato(dato)
            regs.append(reg)
            
        return regs

    #------------------------------------------------------------------------------------------
    def RegistrosPorValores(self, dic_prop_valor):

        regs = []
        for key in self._datos.keys():
            for dato in self._datos[key]:
                reg = self.RegistroUsandoDato(dato)
                reg_cumple = True
                for (k, v) in dic_prop_valor.items():
                    if reg.getv(k) != v:
                        reg_cumple = False
                        break
                if reg_cumple:
                    regs.append(reg)
            
        return regs

    #------------------------------------------------------------------------------------------
    def RegistroUsandoDato(self, dato):

        reg = Registro()
        c = 0
        for h in self._header:
            reg.setv(h, dato[c] if c < len(dato) else '')
            c += 1
        return reg

    #------------------------------------------------------------------------------------------
    def ContarRegistros(self):

        r = self._archivo.CantidadLineas()
        return 0 if r == 0 else r - 1 #restar header si hay datos

    #------------------------------------------------------------------------------------------
    def LeerRegistros(self):

        info = {'reg_repetidos':[], 'reg_erroneos':[], 'info_repetidos':[], 'info_erroneos':[], }

        if not self._archivo.Existe():
            IOSistema().PrintLine('[ %s ] Error: El archivo no existe' % (self._archivo))
            exit()

        self._datos.clear()
        self._archivo.RepararEncoding()

        progreso = Progreso(5, self.ContarRegistros(), '[ %s ] Leyendo' % (self._archivo))

        with IOLector(self._archivo).Abrir(binario=True) as ior:
        
            self._reader = csv.reader(ior.Stream(), delimiter=self._delim, quotechar=self._quote)
    
            for hdr in self._reader:
                self._header = Encoding().NormalizarLista([h.lower() for h in hdr]) 
                break
    
            for dato in self._reader:
                try:
                    progreso.Incrementar()

                    reg = self.RegistroUsandoDato(dato)
                    key = self._func_reg_id(reg)
                    if key not in self._datos:
                        self._datos[key] = []
                    else:
                        info['reg_repetidos'].append(reg)
                        info['info_repetidos'].append(key)
                    self._datos[key].append(Encoding().NormalizarLista(dato))
                    
                except Exception as e:
                    info['reg_erroneos'].append(reg)
                    info['info_erroneos'].append(e)
    
        self._inicializado = True
        
        return info

#=============================================================================================
class CSVEscritor:
    
    #------------------------------------------------------------------------------------------
    def __init__(self, archivo_csv_output='nul', headerstruc=None, escribir_header=True, delim=',', quote='"', quoting=csv.QUOTE_ALL):
        
        self._nombre = archivo_csv_output
        self._archivo = archivo_csv_output
        self._headerstruc = headerstruc
        self._escribir_header = escribir_header
        self._delim = delim
        self._quote = quote
        self._quoting = quoting

        if archivo_csv_output.Existe():
            IOSistema().PrintLine('[ %s ] Warning: El archivo ya existe' % (archivo_csv_output))
        if not archivo_csv_output.CarpetaPadre().Existe():
            IOSistema().PrintLine('[ %s ] Warning: La carpeta no existe, creando..' % (archivo_csv_output))
            archivo_csv_output.CarpetaPadre().Crear()

        self._writer = csv.writer(open(archivo_csv_output.Ruta(), 'wb'), delimiter=self._delim, quotechar=self._quote, quoting=self._quoting)
        
        if self._escribir_header:
            self.EscribirHeader()
        
    #------------------------------------------------------------------------------------------
    def Nombre(self):
        return self._nombre

    #------------------------------------------------------------------------------------------
    def Archivo(self):
        return self._archivo

    #------------------------------------------------------------------------------------------
    def EscribirHeader(self):

        if self._headerstruc is None:
            raise RuntimeError('[ %s ] Error: El _headerstruc no ha sido definido' % (self._archivo))
        
        header = []
        for tdato in self._headerstruc:
            header.append(tdato.Nombre().lower())

        header = Encoding().NormalizarLista(header)
        self._writer.writerow(header)

    #------------------------------------------------------------------------------------------
    def EscribirDatos(self, datos):
        
        progreso = Progreso(5, len(datos), '[ %s ] Escribiendo' % (self._archivo))

        for dato in datos:
            progreso.Incrementar()
            dato = Encoding().NormalizarLista(dato)
            self._writer.writerow(dato)
        
    #------------------------------------------------------------------------------------------
    def EscribirRegistros(self, regs):
        
        progreso = Progreso(5, len(regs), '[ %s ] Escribiendo' % (self._archivo))
        
        for reg in regs:
            progreso.Incrementar()
            self.EscribirRegistro(reg)

    #------------------------------------------------------------------------------------------
    def EscribirRegistro(self, reg):
        
        if self._headerstruc is None:
            raise RuntimeError('[ %s ] Error: El _headerstruc no ha sido definido' % (self._archivo))

        dato = []
        for tdato in self._headerstruc:
            header_nombre = tdato.Nombre()
            dato.append(reg.getv(header_nombre))
            
        self._writer.writerow(dato)
        
    #------------------------------------------------------------------------------------------
    def RedirigirSalida(self, archivo_csv_output):
        
        self.__init__(archivo_csv_output, self._headerstruc, self._escribir_header, self._delim, self._quote, self._quoting)

    #------------------------------------------------------------------------------------------
    def RegistroNuevo(self):
        reg = Registro()
        for tdato in self._headerstruc:
            reg.setv(tdato.Nombre(), '')
        return reg




import sys

from PyQt4 import QtGui, Qt, QtCore
from horno.utiles.IO import IOSistema
from horno.utiles.Singleton import Singleton


#=============================================================================================
class UIControles (metaclass=Singleton):
   
    #------------------------------------------------------------------------------------------
    def __init__(self):
        
        ''

    #------------------------------------------------------------------------------------------
    def AgregarControles(self, ctl_padre, layout_padre, info_ctl_hijos, txt_padre=None, width_label=100, horz=True):
        
        if horz:
            layout_hijo = QtGui.QHBoxLayout()
        else:
            layout_hijo = QtGui.QVBoxLayout()
            
        layout_padre.addLayout(layout_hijo)

        if txt_padre:
            lbl_padre = QtGui.QLabel(txt_padre, ctl_padre)
            if width_label: lbl_padre.setFixedWidth(width_label)
            layout_hijo.addWidget(lbl_padre)

        for (txt_hijo, ctl_hijo) in info_ctl_hijos:
            if txt_hijo:
                lbl_hijo = QtGui.QLabel(txt_hijo, ctl_padre)
                if width_label: lbl_hijo.setFixedWidth(width_label)
                layout_hijo.addWidget(lbl_hijo)
            layout_hijo.addWidget(ctl_hijo)

    #------------------------------------------------------------------------------------------
    def GenerarControlMultiple(self, items_dict, items_checked=[], width=0, height=0):

        ctl_lista = Qt.QListView()
        model = Qt.QStandardItemModel(ctl_lista)

        items = sorted(items_dict.items(), key = lambda p: p[0])
        for (clave, valor) in items: 
            ctl_item = Qt.QStandardItem('%s | %s' % (clave, str(valor)))
            ctl_item.setFlags(Qt.Qt.ItemIsUserCheckable | Qt.Qt.ItemIsEnabled)
            ctl_item.setData(Qt.Qt.Checked if clave in items_checked else Qt.Qt.Unchecked, Qt.Qt.CheckStateRole)
            model.appendRow(ctl_item)
        
        ctl_lista.setModel(model)
        if width: ctl_lista.setFixedWidth(width)
        if height: ctl_lista.setFixedHeight(height)
        
        return ctl_lista      

    #------------------------------------------------------------------------------------------
    def MostrarMensaje(self, mensaje, tipo):
        
        if tipo == 'ERROR':
            mb = QtGui.QMessageBox(QtGui.QMessageBox.Critical, tipo, mensaje)
        elif tipo == 'WARNING':
            mb = QtGui.QMessageBox(QtGui.QMessageBox.Warning, tipo, mensaje)
        elif tipo == 'INFO':
            mb = QtGui.QMessageBox(QtGui.QMessageBox.Information, tipo, mensaje)
            
        mb.exec_()

#=============================================================================================
class UISalida ():
    
    #------------------------------------------------------------------------------------------
    def __init__(self, ui, ctl):
        
        self.ui = ui
        self.ctl = ctl

        self.ui.connect(self.ui, QtCore.SIGNAL('Print'), self.PrintUi)
    
    #------------------------------------------------------------------------------------------
    def PrintLine(self, texto):
        
        self.Print(texto + IOSistema().NewLine())

    #------------------------------------------------------------------------------------------
    def Print(self, texto):
        
        self.ui.emit(QtCore.SIGNAL('Print'), texto)

    #------------------------------------------------------------------------------------------
    def PrintUi(self, texto):
        
        texto_ant = str(self.ctl.toPlainText()).encode(sys.stdout.encoding, errors='replace')
        self.ctl.setText(texto_ant + texto)
        self.ctl.moveCursor(QtGui.QTextCursor.End)




from horno.datos.Encoding import Encoding
from horno.utiles.IO import IOSistema
from horno.utiles.Singleton import Singleton
import html
import re


#==============================================================================================
class HtmlHelper (metaclass=Singleton):
    
    #------------------------------------------------------------------------------------------
    def UnescapearTexto(self, texto):
                        
        return html.unescape(texto)
        
    #------------------------------------------------------------------------------------------
    def CargarUrl(self, url):

        from lxml import etree, html
        from urllib import request

        url_partes = url.split('://', 2)
        url = '%s://%s' % (url_partes[0], url_partes[-1].replace('//', '/'))

        req = request.Request(url, headers={'User-Agent' : "Magic Browser"})
        page_content = request.urlopen(req).read()
        page_content = re.sub(b'\<\!\[CDATA\[(.+)\]\]\>', br'\1', page_content)
        page_elem = html.fromstring(page_content)
        
        #parser = etree.XMLParser(strip_cdata=False)
        #page_elem = etree.XML(page_content, parser)
        
        return page_elem

    #------------------------------------------------------------------------------------------
    def ElemToString(self, elem):
        
        from lxml import etree
        
        inner_html = etree.tostring(elem)
        return inner_html        

    #------------------------------------------------------------------------------------------
    def XPathSafeText(self, elem, paths, default=''):
        
        try:
            if type(elem) in [str, bytes] or any([t in str(type(elem)) for t in ['_ElementStringResult', '_ElementUnicodeResult']]): 
                return Encoding().NormalizarTexto(elem)
            if not type(paths) == list: 
                paths = [paths]
            for path in paths:
                res = elem.xpath(path)
                if not res: continue
                txt = Encoding().NormalizarTexto(res[0]).strip()
                if not txt: continue
                return txt
            return default
        except Exception as e:
            IOSistema().PrintLine(e)
            return '[ERROR-xpathtext]'

    #------------------------------------------------------------------------------------------
    def XPathSafeProp(self, elem, path, prop, default=''):

        try:
            res = elem.xpath(path)
            if not res: return default
            txt = Encoding().NormalizarTexto(res[0].get(prop))
            return txt
        except Exception as e:
            IOSistema().PrintLine(e)
            return '[ERROR-xpathprop]'
        


#==============================================================================================
class UrlHelper ():
    
    #------------------------------------------------------------------------------------------
    def __init__(self, url, url_rel=''):
        
        from urllib.parse import urlparse
        
        self.url_data = urlparse(url)
        self.esquema = self.url_data.scheme or 'http'
        self.hostpuerto = self.url_data.netloc or '.'
        self.host = self.url_data.hostname or '.'
        self.ruta = self.url_data.path
        self.puerto = self.url_data.port
        self.params = self.url_data.params
        self.query = self.url_data.query
        
        if url_rel and self.host == '.':
            url_data_rel = urlparse(url_rel)
            self.esquema = url_data_rel.scheme or 'http'
            self.hostpuerto = url_data_rel.netloc or '.'
            self.host = url_data_rel.hostname or '.'
            self.puerto = url_data_rel.port
            if not self.ruta.startswith('/'):
                self.ruta = '%s/%s' % (url_data_rel.path, self.url_data.path)
        
        self.ruta = self.ruta.replace('//', '/') 

    #------------------------------------------------------------------------------------------
    def __str__(self):
        
        return self.full_path()

    #------------------------------------------------------------------------------------------
    def full_path(self):

        url_link_ok = '%s://%s/%s%s' % (
            self.esquema, 
            self.hostpuerto,
            self.ruta,
            '?%s' % (self.query) if self.query else ''
        )
        return url_link_ok

    #------------------------------------------------------------------------------------------
    def prev_path(self):
        
        path = self.full_path()
        path_partes = path.rsplit('/', 1)
        return path_partes[0] if '.' in path_partes[-1] else path



# -*- coding:utf8 -*-
"""
@author: Krzysiek Grzembski (mynthon.net)
@version: 0.0.1
@python: 2.6
@license: free for any use
@warranty: no warranty at all
 
--------------------------------------------------------------------------------
--- MANUAL ---------------------------------------------------------------------
--------------------------------------------------------------------------------
 
C data is returned as string right stripped
N data is returned as float
L data is returned as boolean
D data is returned as string
 
dataset[0] contains iformations about delete flag. If True item is marked as
deleted. 
 
First column in dataset contains info about delete flag, but it has to be
transparent to user! For user data always begins with index 0 not 1!
 
--------------------------------------------------------------------------------
--- EXAMPLES -------------------------------------------------------------------
--------------------------------------------------------------------------------
 
conn = connect('file.dbf' [,cols])
Opens database. If second argument is givent creates new empty database.
Second (optional) argument is list of columns containing name and definiton.
Currently only few types are supported.
Numeric column - list of four values: name, type, length, decimal
Character column - list of three values: name, type, length
Date column - list of three values: name, type, length
Logical column - list of three values: name, type, length
 
example of creating new database:
 
columns = (
    ('MYNUM', 'N', 8, 2),
    ('MYCHAR', 'C', 30),
    ('MYLOGICL', 'L'),
    ('MYDATE', 'D')
)
 
dbu = connect('mydb.dbf', columns)
 
cursor = conn.cursor()
 
cursor.description
# return tuple containing 7-items tuples. First item for every tuple is column
# name and other items are None.
# (
#   (colname1, None, None, None, None, None, None),
#   (colname2, None, None, None, None, None, None),
#   (colname3, None, None, None, None, None, None),
# )
 
cursor.descriptionDbf
# Returns tuple containing column names.
# (colname1, colname2, colname3 ..., colnameN)
 
cursor.descriptionDbf2
# returns tuple containing tuples for every column. Data in every tuple is
# column name, field length, field type and decimal places.
# (
#   (colname1, length, type, decimal),
#   (colname2, length, type, decimal),
#   (colname3, length, type, decimal),
# )
 
conn.commit('newfile.dbf')
# Saves file. If no name is given saves changes to current file.
 
LINKS:
http://code.activestate.com/recipes/362715/
http://www.whitetown.com/dbf-format/dbf.php
http://www.clicketyclick.dk/databases/xbase/format/
http://www.clicketyclick.dk/databases/xbase/format/dbf.html#DBF_STRUCT
"""
 
import time
import sys
import math
import struct
import re
import decimal
 
class DbfOnPyError(Exception):
    pass
 
class DbfOnPyDataError(Exception):
    pass
 
class DbfOnPySqlError(Exception):
    pass
 
def connect(f, cols=None):
    return Connection(f, cols)
 
class Cursor():
    def __init__(self, parent):
        self._parent = parent
        self._cnt = 0
        self._dataset = parent._dataset
        self._recordset = parent._recordset
        self.description = None
        self.descriptionDbf = None
        self.descriptionDbf2 = None
        self._description()
        self.rowcount = len(self._dataset)
     
    def _description(self):
        set = [] # py db api
        set2 = [] # colnames only
        set3 = [] # column full info (name, len, type, decimal)
        set.append(('MARKED_FOR_DELETE',None,None,None,None,None,None))
        set2.append('MARKED_FOR_DELETE')
        set3.append(('MARKED_FOR_DELETE', 1, '', 0))
        for record in self._recordset:
            set.append((record['name'],None,None,None,None,None,None))
            set2.append(record['name'])
            set3.append((
                record['name'],
                record['length'],
                record['type'],
                record['decimal']
            ))
        self.description = tuple(set)
        self.descriptionDbf = tuple(set2)
        self.descriptionDbf2 = tuple(set3)
     
    def __analyzeSqlQuery(self, query):
        """\
        Method analyzes sql query and overwrites selecting methods.
        Methods: fetchone(), fetchall(), next() are overwritten at runtime.
        It allows to spped up selecting of data.
        """
        sqlQuery = query.lower().strip().rstrip(';').rstrip()
        sqlQuery = re.sub('\s+', ' ', sqlQuery)
         
        if sqlQuery == 'delete from dbf':
            self.zap()
             
        elif sqlQuery == 'commit':
            self.self.commit()
             
        elif sqlQuery == 'select * from dbf' or sqlQuery == '':
            # fetch all records (default)
            def next():
                if self._cnt >= len(self._dataset):
                    self._cnt = 0
                    raise StopIteration
                else:
                    self._cnt += 1
                    return self._dataset[self._cnt - 1]
            self.next = next
             
            def fetchall():
                return self._dataset
            self.fetchall = fetchall
             
            def fetchone():
                if self._cnt >= len(self._dataset):
                    return None
                else:
                    self._cnt += 1
                    return self._dataset[self._cnt - 1]
            self.fetchone = fetchone
             
        elif sqlQuery == 'select * from dbf where delete_flag = true':
            # fetch all deleted rows
            def next():
                while 1:
                    if self._cnt >= len(self._dataset):
                        self._cnt = 0
                        raise StopIteration
                    else:
                        self._cnt += 1
                        if self._dataset[self._cnt - 1][0] == True:
                            return self._dataset[self._cnt - 1]
            self.next = next
             
            def fetchall():
                return [row for row in self._dataset if row[0] == True]
            self.fetchall = fetchall
             
            def fetchone():
                while 1:
                    if self._cnt >= len(self._dataset):
                        return None
                    else:
                        self._cnt += 1
                        if self._dataset[self._cnt - 1][0] == True:
                            return self._dataset[self._cnt - 1]
            self.fetchone = fetchone
             
        elif sqlQuery == 'select * from dbf where delete_flag = false':
            # fetch all not deleted rows
            def next():
                while 1:
                    if self._cnt >= len(self._dataset):
                        self._cnt = 0
                        raise StopIteration
                    else:
                        self._cnt += 1
                        if self._dataset[self._cnt - 1][0] == False:
                            return self._dataset[self._cnt - 1]
            self.next = next
             
            def fetchall():
                return [row for row in self._dataset if row[0] == False]
            self.fetchall = fetchall
             
            def fetchone():
                while 1:
                    if self._cnt >= len(self._dataset):
                        return None
                    else:
                        self._cnt += 1
                        if self._dataset[self._cnt - 1][0] == False:
                            return self._dataset[self._cnt - 1]
            self.fetchone = fetchone
             
        else:
            raise DbfOnPySqlError(
                '\nWrong query:\n'\
                '%s\n'\
                'Allowed are:\n'\
                'EMPTY STRING\n'\
                'select * from dbf\n'\
                'select * from dbf where delete_flag = true\n'\
                'select * from dbf where delete_flag = false\n' % sqlQuery
            )
     
    def fetchone(self):
        return None
     
    def fetchall(self):
        return []
     
    def rowcount(self):
        return len(self._dataset)
     
    def execute(self, sql='', data=None):
        self.__analyzeSqlQuery(sql)
        return True
     
    def __iter__(self):
        return self
     
    def next(self):
        raise StopIteration
     
    # *********************
    # dbf specific options
    # *********************
    def delete(self):
        """ Mark current record as deleted """
        self._dataset[self._cnt - 1][0] = True
     
    def undelete(self):
        """ Mark current record as not deleted """
        self._dataset[self._cnt - 1][0] = False
     
    def zap(self):
        """ Marks all records as deleted """
        for i in range(len(self._dataset)):
            self._dataset[i][0] = True
     
    def pack(self):
        """ Remove all records marked as deleted. """
        # go from end to begin to avoid reindexing issues
        i = len(self._dataset) - 1
        while i >= 0:
            if self._dataset[i][0]:
                del self._dataset[i]
            i -= 1
     
    def toTxt(self, filename, delimiter=',', lineEnd='\r\n', header=False):
        """Exports current selection to text file."""
        fp = open(filename, 'wb')
         
        headerTxt = ''
        if header:
            headerData = []
            for record in self._recordset:
                headerData.append(record['name'])
            headerTxt = delimiter.join(headerData) + lineEnd
        fp.write(headerTxt)
         
        lineFormat = delimiter.join(
            ['%s'] * len(self._recordset)
            ) + lineEnd
         
        for line in self:
            fp.write(lineFormat % tuple(line[1:]))
         
        fp.close()
     
    def toSqlite3(self, filename):
        pass
     
    def insert(self, data):
        """
        None insert(data)
        inserts new data. data length has to be equal to number of columns
        (delete flag not included - it is always False for new records)
        and must have right order. File is not saved automatically!  
        """
        if len(data) == len(self._recordset):
            i = 0
             
            # prepare record to insert
            while i < len(self._recordset):
                if self._recordset[i]['type'] == 'C':
                    data[i] = self._parent._format_C_put_(data[i], self._recordset[i]['length'])
                elif self._recordset[i]['type'] == 'N':
                    data[i] = self.__fmt_N_upd(data[i], self._recordset[i]['length'], self._recordset[i]['decimal'])
                elif self._recordset[i]['type'] == 'L':
                    data[i] = self.__fmt_L_upd(data[i])
                else:
                    #domyslnie traktuj jak string
                    data[i] = self._parent._format_C_put_(data[i], self._recordset[i]['length'])
                i += 1
             
            # add delete flag to record
            data.insert(0, False)
            # add record
            self._dataset.append(data)
    ### /insert
     
    def update(self, data):
        """ modify current record """
        if len(data) == len(self._recordset):
            i = 0
            while i < len(self._recordset):
                if self._recordset[i]['type'] == 'C':
                    self._dataset[self._cnt - 1][i+1] = self._parent._format_C_put_(data[i], self._recordset[i]['length'])
                elif self._recordset[i]['type'] == 'N':
                    self._dataset[self._cnt - 1][i+1] = self.__fmt_N_upd(data[i], self._recordset[i]['length'], self._recordset[i]['decimal'])
                elif self._recordset[i]['type'] == 'L':
                    self._dataset[self._cnt - 1][i+1] = self.__fmt_L_upd(data[i])
                else:
                    #domyslnie traktuj jak string
                    self._dataset[self._cnt - 1] = self._parent._format_C_put_(data[i], self._recordset[i]['length'])
                i += 1
    ### /update
     
    #########################################################
    # section: datachecking
    # check data for insert and update operations
    #########################################################
    def __fmt_L_upd(self, da):
        if type(da) is bool or da == ' ':
            return da
        raise DbfOnPyDataError('Invalid value for BOOL column: "%s"' % da)
     
    def __fmt_N_upd(self, num, numLen, decPlaces):
        """Returns number if is allowed."""
         
        """
        In dbf length of numeric field is total length with decimal separator
        and minus sign so number -174.976 has length 8.
         
        In field of length 5 and 2 decimal places you can put numbers from
        -9.99 to 99.99.
         
        My idea to check if number fits field is to multiple number by
        10 to power "decimalPlaces". Eg. (length=6, decimal=2) number 853.8776
        become 85387.76. Now i round it to get 85388.
         
        This number cannot be bigger than 10^fieldLength if there is no decimal
        separator and 10^(fieldLength-1) if there is decimal separator (thats
        because decimal separator takes one slot from total length).
         
        I my example length=6 and there is decimal separator (decimalPlaces>0) so
        biggest number is 10^(6-1)=100000. 85388 is smaller than 100000 so it is ok.
         
        For the same field number 999.99998 is not ok because:
        999.99998 * 10^decimalPlaces = 99999.998
        after rounding i get 100000. 100000 == 10000 so number doesnt fit.
         
        For the same field number 1111.5 is not ok because:
        1111.5 * 10^decimalPlaces = 111150
        after rounding i get 111150. 111150 > 10000 so number doesnt fit.
        """
         
        point = 1 - 0 ** decPlaces
         
        number = decimal.Decimal(str(num))
        if number.is_zero():
            return decimal.Decimal('0')
         
        number2 = (number * 10 ** (decPlaces)).quantize(decimal.Decimal('1')) 
        #print number, number2, (10 ** (numLen - 1)), point
         
        if ((10 ** (numLen - point)) > number2 > (10 ** (numLen - point - 1) * (-1))):
            return number
        else:
            raise DbfOnPyDataError('After rounding number %s does not fit to cell of lenght %s with %s decimal places.' % (num, numLen, decPlaces))
            return None
 
    def close(self):
        'agregado por AFS'
 
class Connection:
    #########################################################
    # section: basic methods
    #########################################################
     
    def __init__(self, fname, cols):
         
        # jesli jest cols to znaczy ze masz stworzyc nowa pusta baze danych
        if cols != None:
            self.__createEmptyDB(fname, cols)
         
        # te zmienne sluza do porownan, zeby nie tworzyc obiektow
        # w kazdym IF...ELSE czy RETURN
        self.DATA_DBF_L_TRUE = 'T'
        self.DATA_DBF_L_FALSE = 'F'
        self.DATA_DBF_L_EMPTY = ' '
         
        self.DATA_DBF_DELFLAG_TRUE = '\x2a' # '*'
        self.DATA_DBF_DELFLAG_FALSE = '\x20' # ' '
         
        # inne
        self.currentFileName = ''
        self.header = {}
        self._recordset = []
        self._dataset = []
        self.idx = [{}] # indexes
         
        self.currentFileName = fname;
         
        f = open(fname, 'rb')
         
        '''# header
        self.header = {
            'fileType' : f.read(1),
            'lastUpdate' : f.read(3), # YMD
            'numberOfRecords' : self._byteToInt_(f.read(4)),
            'positionFirstData' : f.read(2), # where first data starts (bytes)
            'lengthOfData' : self._byteToInt_(f.read(2)), # length of one data record with del flag
            'reserved1' : f.read(16),
            'tableFlag' : f.read(1),
            'codePage' : f.read(1),
            'reserved2' : f.read(2)
        } '''
         
        (
        self.header['fileType'],
        self.header['lastUpdate'], # YMD
        self.header['numberOfRecords'],
        self.header['positionFirstData'], # where first data starts (bytes)
        self.header['lengthOfData'],# length of one data record with del flag
        self.header['reserved1'],
        self.header['tableFlag'],
        self.header['codePage'],
        self.header['reserved2']
        ) = struct.unpack('<c3sIHH16scc2s', f.read(32))
         
        # record set
        while 1:
            if f.read(1) == '\x0d':
                f.seek(-1,1) # after checking for delimiter go back 1 char
                break
            else:
                f.seek(-1,1) # after checking for delimiter go back 1 char
                self._recordset.append(
                    {
                        'name' : self._format_COLNAME_get_(f.read(10)),
                        'break' : f.read(1),
                        'type' : f.read(1),
                        'displacement' : f.read(4),
                        'length' : self._byteToInt_(f.read(1)),
                        'decimal' : self._byteToInt_(f.read(1)),
                        'flag' : f.read(4),
                        'autoincrement' : f.read(1),
                        'step' : f.read(1),
                        'reserved' : f.read(8)
                    }
                )
         
        # skip terminator 0x0d and database container 0x00
        f.read(1)
        while 1:
            if f.read(1) != '\x00':
                f.seek(-1,1)
                break
         
        # data
         
        ##### PREKOMPILACJA #######
         
        # tutaj przygotowujemy i prekompilujemy funkcje, dzieki czemu
        # przy duzych plikach przyspieszamy wczytywanie do 50%!
        dataAppendConf = []
        i = 1 # @ 0 is delete flag
         
        for j in self._recordset:
            le = j['length']
            if j['type'] == 'C':
                dataAppendConf.append('strline[%s:%s].rstrip()' % (i, i+le))
            elif j['type'] == 'N':
                dataAppendConf.append('self._format_N_get_(strline[%s:%s])' % (i, i+le))
            elif j['type'] == 'L':
                dataAppendConf.append('self._format_L_get_(strline[%s:%s])' % (i, i+le))
            else:
                dataAppendConf.append('strline[%s:%s]' % (i, i+le))
            i += le
         
        compilestr  = ''\
            '\ndef appendfuncdynamic(self, strline):'\
            '\n    temp = False;'\
            '\n    if strline[0] == self.DATA_DBF_DELFLAG_TRUE: temp = True;'\
            '\n    self._dataset.append([temp, %s])' % ','.join(dataAppendConf)
         
        exec(compilestr)
         
        while 1:
            dbfLine = f.read(self.header['lengthOfData'])
            if len(dbfLine) != self.header['lengthOfData'] or dbfLine[0] == '\x1a':
                break
            else:
                eval('appendfuncdynamic(self, dbfLine)')
        f.close()
         
    ### /__init__()
     
    def __createEmptyDB(self, fname, cols):
        """ Create empty database """
         
        # ustaw dane do naglowka
        self.header = {
            'fileType' : '\x03',
            'lastUpdate' : self._lastUpdate_(time.strftime("%y%m%d")), # YMD
            'numberOfRecords' : 0,
            'positionFirstData' : 0, # where first data starts (bytes)
            'lengthOfData' : 0, # length of one data record with del flag
            'reserved1' : '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            'tableFlag' : '\x00',
            'codePage' : '\x00',
            'reserved2' : '\x00\x00'
        }
         
        # clear current recordset info (columns definitions)...
        self._recordset = []
         
        # ... and add new one.
        for row in cols:
            # ('name', 'type', length, decimal)
            ctypes = ('L','N','C','D')
            cname = '%s%s' % (row[0].upper(), (10 - len(row[0])) * '\x00')
            clen = 1
            cdec = 0
            ctype = row[1].upper()
             
            if ctype not in ctypes:
                ctype = 'C'
             
            try:
                if ctype == 'N':
                    cdec = row[3]
            except:
                pass
             
            try:
                if ctype == 'C' or ctype == 'N':
                    clen = row[2]
                elif ctype == 'D':
                    clen = 8
                elif ctype == 'L':
                    clen = 1
            except:
                pass
             
            self._recordset.append(
                {
                    'name' : cname,
                    'break' : '\x00',
                    'type' : ctype,
                    'displacement' : '\xb8\x60\x0d\x00',
                    'length' : clen,
                    'decimal' : cdec,
                    'flag' : '\xec\x75\x02\x65',
                    'autoincrement' : '\xaa',
                    'step' : '\x21',
                    'reserved' : '\x72\x4f\x0d\x00\xec\x75\x12\x65'
                }
            )
        # wyczysc stare dane
        self._dataset = [] 
        self.commit(fname)
    ### /__createEmptyDB
     
    def commit(self,f=''):
        if f == '':
            f = self.currentFileName
        dbf = open(f, "wb")
         
        # recordset
        dataFieldLength = 1 #delete flag
        _recordset = b''
        for i in range(len(self._recordset)):
            _recordset += self._format_COLNAME_put_(self._recordset[i]['name'])
            _recordset += self._recordset[i]['break']
            _recordset += self._recordset[i]['type']
            _recordset += self._recordset[i]['displacement']
            _recordset += self._intToByte_(self._recordset[i]['length'], 1)
            _recordset += self._intToByte_(self._recordset[i]['decimal'], 1)
            _recordset += self._recordset[i]['flag']
            _recordset += self._recordset[i]['autoincrement']
            _recordset += self._recordset[i]['step']
            _recordset += self._recordset[i]['reserved']
            dataFieldLength += self._recordset[i]['length']
         
        # terminator and db container
        _TerminatorAndDBContainer = b'\x0d\x00'
         
        # header
        '''
        _header = ''
        _header += self.header['fileType']
        _header += self._lastUpdate_(time.strftime("%y%m%d"))# last update
        _header += self._intToByte_(len(self._dataset),4) # num of records
        _header += self._intToByte_(32 + len(_TerminatorAndDBContainer) + len(_recordset), 2) # 
        _header += self._intToByte_(dataFieldLength,2) # length of data record
        _header += self.header['reserved1']
        _header += self.header['tableFlag']
        _header += self.header['codePage']
        _header += self.header['reserved2']
        '''
         
        _header = struct.pack('<c3sIHH16scc2s', 
            self.header['fileType'],
            self._lastUpdate_(time.strftime("%y%m%d")),
            len(self._dataset),
            32 + len(_TerminatorAndDBContainer) + len(_recordset),
            dataFieldLength,
            self.header['reserved1'],
            self.header['tableFlag'],
            self.header['codePage'],
            self.header['reserved2']
        )
         
        # save header
        dbf.write(_header + _recordset + _TerminatorAndDBContainer)
         
        # save data        
        for rowdata in self._dataset:
            _data = ''
            columnNo = 0
            for field in rowdata:
                if columnNo == 0:
                    _data += self.__format_DELETEFLAG_put(field)
                else:
                    if self._recordset[columnNo-1]['type'] == 'N':
                        _data += self._format_N_put_(
                            field,
                            self._recordset[columnNo-1]['length'],
                            self._recordset[columnNo-1]['decimal']
                        )
                    elif self._recordset[columnNo-1]['type'] == 'L':
                        _data += self._format_L_put_(field)
                    else:
                        _data += self._format_C_put_(
                            field,
                            self._recordset[columnNo-1]['length']
                        )
                columnNo += 1
            dbf.write(_data)
        #end of file     
         
        dbf.write('\x1a')
        dbf.close()
    ### /commit()
     
    #########################################################
    # section: cursor
    #########################################################
     
    def cursor(self):
        """ create cursor which contains all rows """
        return Cursor(self)
     
    #########################################################
    # section: data formatting methods
    # _put_ methods translates data from column format to file
    # _get_ methods translates data from file to column format
    #########################################################
     
    def _format_C_put_(self, dataString, dataLength):
        if len(dataString) > dataLength:
            raise DbfOnPyError( 'String is too long: ' + \
                str(len(dataString)) + \
                " (max: " + str(dataLength) +" expected)")
        dataString = ("% -" + str(dataLength) + "s") % dataString
        return dataString
     
    def _format_C_get_(self, dataString):
        return dataString.rstrip()
     
    def _format_N_put_(self, dataString, dataLength, decimalPlaces):
        return ("% " + str(dataLength) + "s") % (("%.0" + str(decimalPlaces) + "f") % decimal.Decimal((dataString)))
     
    def _format_N_get_(self, dataString):
        try:
            return decimal.Decimal(dataString)
        except:
            return 0
     
    def _format_L_put_(self, dataString):
        if dataString == True:
            return self.DATA_DBF_L_TRUE
        elif dataString == False:
            return self.DATA_DBF_L_FALSE
        else:
            return self.DATA_DBF_L_EMPTY
     
    def _format_L_get_(self, dataString):
        if dataString == self.DATA_DBF_L_FALSE:
            return False
        elif dataString == self.DATA_DBF_L_TRUE:
            return True
        else:
            return dataString
         
    def _format_D_put_(self):
        pass
     
    def _format_D_get_(self):
        pass
     
    def _format_COLNAME_get_(self, colname):
        return colname[0:colname.find('\x00')]
 
    def _format_COLNAME_put_(self, dataString):
        if len(dataString) > 10:
            return dataString[0:10]
        else:
            return dataString + ((10 - len(dataString)) * '\x00')
 
    def __format_DELETEFLAG_get(self, dataString):
        return (dataString == self.DATA_DBF_DELFLAG_TRUE and True) or False
     
    def __format_DELETEFLAG_put(self, flag):
        return (flag and self.DATA_DBF_DELFLAG_TRUE) or self.DATA_DBF_DELFLAG_FALSE
     
    #########################################################
    # section: helper methods
    #########################################################
     
    def _intToByte_(self, n, l):
        """ Converts integer to bytes. If length aftrer conversion
        is smaller than given as param returned value is right-filled
        with 0x00 bytes. Use Little-endian  byte order."""
        return b''.join([
            chr((n >> ((l - i - 1) * 8)) % 256) for i in range(l)
        ][::-1])
     
    def _byteToInt_(self, bytes):
        """ Converts byte string to integer.
        struct.unpack is too much platform dependent.
        Use Little-endian  byte order."""
        return sum([
            ord(b) << (8 * i) for i, b in enumerate(bytes)
        ])
     
    def _lastUpdate_(self, dateYMD = None):
        """
        _lastUpdate_([str data_YYMMDD]) -> void
        aktualizuje informacje o dacie ostatniej modyfikacji jesli ten parametr
        zostal podany, w przeciwnym wypadku wyswietla date ostatniej modyfikacji.
        """
        if dateYMD == None:
            out = "%02d" % ord(self.header["lastUpdate"][0])
            out += "%02d" % ord(self.header["lastUpdate"][1])
            out += "%02d" % ord(self.header["lastUpdate"][2])
            return out
        else:
            out  = chr(int(dateYMD[0:2]))
            out += chr(int(dateYMD[2:4]))
            out += chr(int(dateYMD[4:6]))
            return out
     
    def _numberOfRecords_(self, records = -1):
        """
        _numberOfRecords_([int ile_rekordow]) -> void
        aktualizuje informacje o ilosci rekordow jesli parametr zostal podany,
        w przeciwnym wypadku przesyla do konsoli informacje o aktualnej ilosci rekordow.
        """
        if records < 0:
            print(self._byteToInt_(self.header['numberOfRecords']))
        else:
            self.header["numberOfRecords"] = self._intToByte_(records,4)
     
    def _lengthOfData_(self, length = -1):
        if length < 0:
            print(self._byteToInt_(self.header['lengthOfData']))
        else:
            self.header["lengthOfData"] = self._intToByte_(length,2)
     
    def _positionFirstData_(self, position = -1):
        if position < 0:
            print(self._byteToInt_(self.header['positionFirstData']))
        else:
            self.header["positionFirstData"] = self._intToByte_(position,2)   

    def close(self):
        'agregado por AFS'



import setuptools

import horno

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=horno.name,
    version=horno.version,
    author="afs",
    author_email="author@example.com",
    description="Lo que el cuerpo necesita",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'python-dateutil', 'bcrypt', 'passlib', 'unidecode', # 'lxml',
    ],                 
)